package transfer

import (
	"context"
	"fmt"
	"log"
	"sync"
	"time"

	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpclient"
	"github.com/it-novum/nsta/nstautil"
	"github.com/it-novum/nsta/oitcdb"
	"github.com/it-novum/nsta/satstatus"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

// PullSatellite is a satellite pulled via http requests from master
type PullSatellite struct {
	SatelliteOptions

	wg sync.WaitGroup

	httpClient *httpclient.NSTAHTTPClient
	lastExport time.Time
	toSat      gearman.SyncMessageFIFO
	shutdown   chan struct{}
}

// SatelliteID returns the db id of the satellite
func (s *PullSatellite) SatelliteID() int {
	return s.DB.ID
}

func (s *PullSatellite) SatelliteName() string {
	return s.DB.Name
}

// APIKey returns the db signature api key
func (s *PullSatellite) APIKey() string {
	return s.DB.APIKey
}

func (s *PullSatellite) LastExport() time.Time {
	return s.lastExport
}

func (s *PullSatellite) SendMessage(msg *gearman.Message) {
	s.toSat.Push(msg)
}

func (s *PullSatellite) run(ctx context.Context) {
	var timeoutCheckSec int
	timeoutDBSec := s.DB.Timeout
	if timeoutDBSec < 240 {
		timeoutCheckSec = 60
	} else if timeoutDBSec < 3600 {
		timeoutCheckSec = timeoutDBSec / 4
	} else {
		timeoutCheckSec = 3600
	}

	ts := time.NewTicker(time.Second * time.Duration(timeoutCheckSec))   // Satellite status update interval
	ticker := time.NewTicker(time.Second * time.Duration(s.DB.Interval)) // Pull check results interval

	nextTs := time.Time{}

	defer func() {
		log.Println(&Error{Name: s.DB.Name, Message: "shutting down"})
		ticker.Stop()
		ts.Stop()
		log.Println(&Error{Name: s.DB.Name, Message: "close message fifo"})
		s.toSat.Close()
		log.Println(&Error{Name: s.DB.Name, Message: "shutdown complete"})
		s.wg.Done()
	}()

	lastSeen := time.Time{}

	for {
		select {
		case <-ctx.Done():
			return
		case _, more := <-s.shutdown:
			if !more {
				return
			}
		case <-ts.C:
			absTimeout := lastSeen.Add(time.Second * time.Duration(timeoutDBSec))
			if !absTimeout.After(time.Now()) {
				log.Println(&Error{Name: s.DB.Name, Message: fmt.Sprintf("Last Seen: %s, Timeout: %s", lastSeen.Format(time.RFC3339), absTimeout.Format(time.RFC3339))})
				s.StatusTracker.Update <- &satstatus.Status{
					SatelliteID: s.DB.ID,
					Status:      oitcdb.SatelliteStatusFailed,
					Err:         &Error{Name: s.DB.Name, Message: "timeout reached"},
				}
			}
		case <-ticker.C:
			//Pull check results form SAT
			status := &satstatus.Status{
				SatelliteID: s.DB.ID,
				Status:      oitcdb.SatelliteStatusOK,
			}
			lastExport := s.LastExport()
			msgs := s.toSat.GetAll(8192) // Get messages to send to Satellite from FIFO main
			origMsgs := msgs
			// we do not queue the timesync message, because it should be sent to the client immediatly
			if s.DB.UseTimesync {
				now := time.Now()
				if now.After(nextTs) {
					nextTs = now.Add(time.Minute)
					ts := gearman.TimesyncMessage{
						SatelliteID: s.DB.ID,
						SetTime:     now,
					}
					if data, err := json.Marshal(ts); err != nil {
						log.Println(&Error{Name: s.DB.Name, Message: "could not create timesync message", Err: err})
					} else {
						msgs = append(msgs, &gearman.Message{
							Type:   gearman.MessageTimesync,
							String: string(data),
						})
					}
				}
			}

			response, err := s.httpClient.SendMessages(msgs, lastExport)
			if err != nil {
				status.Err = &Error{Name: s.DB.Name, Message: "could not send/receive messages", Err: err}
				log.Println(status.Err)
				if len(origMsgs) > 0 {
					// Error while sending messages to sat
					// Put all messages back to fifo cache (except time sync messages)
					s.toSat.PushAll(origMsgs, true)
				}
			}
			if response != nil {
				lastSeen = time.Now()
				status.LastSeen = lastSeen
				if response.SatelliteID != s.DB.ID || response.APIKey != s.DB.APIKey {
					log.Println(&Error{Name: s.DB.Name, Message: fmt.Sprint("sent ", len(msgs), " messages")})
					status.Err = &Error{Name: s.DB.Name, Message: "authentication of response failed"}
					log.Println(status.Err)
				} else {
					log.Println(&Error{Name: s.DB.Name, Message: fmt.Sprint("sent ", len(msgs), " and received ", len(response.Messages), " messages")})
					for _, msg := range response.Messages {
						s.ToMaster <- msg
					}
					// That is the timestamp of lastExport on the sattelite compared to the main system?
					if s.LastExport().After(response.LastExport) {
						// An export happend on the main system - push new monitoring configuration to the sattelite
						tData := nstautil.CreateExportTar(s.NagiosPath, s.DB.ID, true)
						// tData = tar.gz as base64 string
						if tData != "" {
							// Second request to the SAT to transfer the new config
							if exportResponse, err := s.httpClient.ExportExchange(lastExport, tData); err != nil {
								log.Println(err)
								status.Err = err
							} else {
								status.LastExport = exportResponse.LastExport
							}
						}
					}
				}
			} else if err != nil {
				status.Status = oitcdb.SatelliteStatusFailed
				log.Println(&Error{Name: s.DB.Name, Message: fmt.Sprint("sent ", len(msgs), " messages")})
			}
			s.StatusTracker.Update <- status
		}
	}
}

func (s *PullSatellite) Close() {
	close(s.shutdown)
	s.wg.Wait()
}

func (s *PullSatellite) Compare(dbSat *oitcdb.Satellite) bool {
	return s.DB.Compare(dbSat)
}

func (s *PullSatellite) Reload(_ context.Context) {
	if !s.DB.NSTASyncInstance {
		log.Println(&Error{Name: s.DB.Name, Message: "skip export"})
		return
	}
	// Set new timestamp for the run func which push new configs to the SAT if required
	s.lastExport = time.Now()
}

func (s *PullSatellite) Start(parent context.Context) error {
	ctx, cancel := context.WithCancel(parent)
	proxy := ""
	if s.DB.UseProxy > 0 {
		proxy = s.DB.ProxyURL
	}

	// Create new httpClient to query the nsta client on the sattelite server
	httpClient, err := httpclient.New(ctx, s.DB.URL, s.DB.ID, s.DB.APIKey, proxy, s.DB.VerifyCertificate)
	if err != nil {
		cancel()
		return err
	}
	s.httpClient = httpClient

	s.lastExport = time.Now()
	s.shutdown = make(chan struct{})

	s.toSat.Start()
	s.wg.Add(1)
	go func() {
		defer s.wg.Done()

		defer cancel()

		s.run(ctx)
	}()

	return nil
}
